@[Link(ldflags: "#{__DIR__}/cre2.o -Wl,--copy-dt-needed-entries `pkg-config --libs re2`")]
lib LibCre2
  type Options = Void*

  fun opt_new = cre2_opt_new : Options
  fun opt_delete = cre2_opt_delete(op : Options) : Nil

  fun opt_posix_syntax = cre2_opt_posix_syntax(op : Options, flag : Bool) : Nil
  fun opt_longest_match = cre2_opt_longest_match(op : Options, flag : Bool) : Nil
  fun opt_log_errors = cre2_opt_log_errors(op : Options, flag : Bool) : Nil
  fun opt_literal = cre2_opt_literal(op : Options, flag : Bool) : Nil
  fun opt_never_nl = cre2_opt_never_nl(op : Options, flag : Bool) : Nil
  fun opt_case_sensitive = cre2_opt_case_sensitive(op : Options, flag : Bool) : Nil
  fun opt_perl_classes = cre2_opt_perl_classes(op : Options, flag : Bool) : Nil
  fun opt_word_boundary = cre2_opt_word_boundary(op : Options, flag : Bool) : Nil
  fun opt_one_line = cre2_opt_one_line(op : Options, flag : Bool) : Nil
  fun opt_dot_nl = cre2_opt_dot_nl(op : Options, flag : Bool) : Nil
  fun opt_encoding = cre2_opt_encoding(op : Options, encoding : Int32) : Nil
  fun opt_max_mem = cre2_opt_max_mem(op : Options, flag : Bool) : Nil

  struct StringPiece
    data : LibC::Char*
    length : Int32
  end

  type CRe2 = Void*

  fun new = cre2_new(pattern : LibC::Char*, patternlen : UInt32, opt : Options) : CRe2
  fun del = cre2_delete(re : CRe2) : Nil
  fun error_code = cre2_error_core(re : CRe2) : Int32
  fun num_capturing_groups = cre2_num_capturing_groups(re : CRe2) : Int32
  fun program_size(re : CRe2) : Int32

  # Invalidated by further re use
  fun error_string = cre2_error_string(re : CRe2) : LibC::Char*
  fun error_arg = cre2_error_arg(re : CRe2, arg : StringPiece*) : Nil

  CRE2_UNANCHORED   = 1
  CRE2_ANCHOR_START = 2
  CRE2_ANCHOR_BOTH  = 3

  fun match = cre2_match(
    re : CRe2,
    text : LibC::Char*,
    textlen : UInt32,
    startpos : UInt32,
    endpos : UInt32,
    anchor : Int32,
    match : StringPiece*,
    nmatch : Int32
  ) : Int32
end

module CRe2
  struct MatchDataLike
    @str : String
    @matches : Pointer(LibCre2::StringPiece)
    @size : Int32

    def initialize(@str, @matches, @size)
    end

    def [](i : Int32) : String
      if i < @size && @matches[i].data != nil
        String.new(Slice.new(@matches[i].data, @matches[i].length))
      else
        ""
      end
    end
  end

  class Regex < ::Regex
    @jit = false
    @re2 : LibCre2::CRe2
    @failed = false
    @anchored = LibCre2::CRE2_UNANCHORED

    def initialize(@pattern : String, @options : Options = Options::None)
      @source = @pattern
      @re = Regex::PCRE2.compile(pattern, pcre2_options(@options)) do |error_message|
        raise Exception.new(error_message)
      end
      # No multiline flag, it's handled on the pattern
      @pattern = "(?m)#{pattern}" if @options & Regex::Options::MULTILINE != 0
      @re2 = LibCre2.new("x", 1, cre2_options(@options))
      begin
        @re2 = LibCre2.new(@pattern, @pattern.size, cre2_options(@options))
      rescue ex : Exception
        @failed = true
      end
    end

    def match(str : String, pos : Int32 = 0, options : Regex::MatchOptions = :none) : MatchData | MatchDataLike | Nil
      msize = LibCre2.num_capturing_groups(@re2) + 1
      if !@failed
        matches = Pointer(LibCre2::StringPiece).malloc(msize)
        if LibCre2.match(@re2, str, str.size, pos, str.size,
             @anchored, matches, msize) != 0
          MatchDataLike.new(str, matches, msize)
        else
          nil
        end
      else
        super
      end
    end

    private def pcre2_options(options : Regex::Options)
      flags = LibPCRE2::UTF | LibPCRE2::DUPNAMES |
              LibPCRE2::UCP
      flags |= LibPCRE2::MULTILINE if options & Regex::Options::MULTILINE != 0
      flags |= LibPCRE2::DOTALL if options & Regex::Options::DOTALL != 0
      flags |= LibPCRE2::CASELESS if options & Regex::Options::IGNORE_CASE != 0
      flags |= LibPCRE2::ANCHORED if options & Regex::Options::ANCHORED != 0
      flags |= LibPCRE2::NO_UTF_CHECK
      flags
    end

    private def cre2_options(options : Regex::Options) : LibCre2::Options
      opts = LibCre2.opt_new
      # Not configurable
      LibCre2.opt_posix_syntax(opts, false)
      LibCre2.opt_longest_match(opts, false)
      LibCre2.opt_log_errors(opts, false)

      LibCre2.opt_case_sensitive(opts, options & Regex::Options::IGNORE_CASE == 0)
      # Not considered when posix_syntax is false
      # LibCre2.opt_perl_classes(opts, options & Regex::Options::PERL_CLASSES != 0)
      # LibCre2.opt_word_boundary(opts, options & Regex::Options::WORD_BOUNDARY != 0)
      # LibCre2.opt_one_line(opts, options & Regex::Options::ONE_LINE != 0)
      LibCre2.opt_dot_nl(opts, options & Regex::Options::DOTALL != 0)
      @anchored = LibCre2::CRE2_ANCHOR_BOTH if options & Regex::Options::ANCHORED != 0
      LibCre2.opt_encoding(opts, 1)
      opts
    end
  end
end

# re = CRe2::Regex.new(".*(x).*", Regex::Options::ANCHORED | Regex::Options::MULTILINE)
# m = re.match("axb")
# p! m.try &.[0]
# p! m.try &.[1]

# re = Regex.new(".*(x).*", Regex::Options::ANCHORED | Regex::Options::MULTILINE)
# m = re.match("axb")
# p! m.try &.[0]
# p! m.try &.[1]

